#include <cstdarg>
#ifndef linux
// In SunOS, sys/ioctl.h doesn't define ioctl()
// function, but it's instead in unistd.h.
#include <unistd.h>
#endif

#include "cons.hh"

#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif

#ifdef DJGPP
#include <conio.h>
#else
#ifdef HAVE_TERMIO_H
#include <termio.h>
#endif
#ifdef HAVE_TERMIOS_H
#include <termios.h>
#endif
#endif

#include "cons_map.hh"

Screen::Screen() : xpos(0), ypos(0), textattr(0xF8), LINES(25), COLS(80)
{
    Allocated = new UsedMap;
    GetScreenGeometry();
}
Screen::~Screen()
{
    Color(0x07);
    delete Allocated;
}

void Screen::GetScreenGeometry()
{
#if defined(DJGPP) || defined(__BORLANDC__)
    struct text_info w;
    gettextinfo(&w);
    LINES=w.screenheight;
    COLS =w.screenwidth;
#else
#ifdef TIOCGWINSZ
    struct winsize w;
    if(ioctl(1, TIOCGWINSZ, &w) >= 0 || ioctl(2, TIOCGWINSZ, &w) >= 0 || ioctl(0, TIOCGWINSZ, &w) >= 0)
    {
        LINES=w.ws_row;
        COLS =w.ws_col;
    }
#else
#ifdef WIOCGETD
    struct uwdata w;
    if(ioctl(1, WIOCGETD, &w) >= 0 || ioctl(2, WIOCGETD, &w) >= 0 || ioctl(0, WIOCGETD, &w) >= 0)
    {
        LINES = w.uw_height / w.uw_vs;
        COLS  = w.uw_width / w.uw_hs;
    }
#endif
#endif
#endif
    Allocated->resize(COLS, LINES);
}

void Screen::allocate(const xycoord &orig, const xycoord &size)
{
    Allocated->SetBox(orig.x, orig.y, size.x, size.y, am_used);
}

void Screen::deallocate(const xycoord &orig, const xycoord &size, bool really)
{
    Allocated->SetBox(orig.x, orig.y, size.x, size.y, really ? am_free : am_bad);
}

Screen::ScreenRegion Screen::AllocateBox(unsigned ysiz, unsigned xsiz, bool temporary)
{
    // fprintf(stderr, "Would like to allocate %ux%u\n", xsiz,ysiz);
    const unsigned ally = Allocated->ysize();
    const unsigned allx = Allocated->xsize();
    
    if(!Allocated->size() || !ysiz || !xsiz)return ScreenRegion();
    unsigned xmin = (xsiz==nsize) ? 1 : xsiz;
    unsigned ymin = (ysiz==nsize) ? 1 : ysiz;
    if(xmin > allx || ymin > ally)return ScreenRegion();
    
    unsigned most = 0;
    xycoord mostpos = {0,0};
    xycoord mostsiz = {0,0};
    bool found=false;
    
    Allocated->Analyze();
    
    // mx,my are the bottomright limits: Required box can not be found further.
    
    for(unsigned y=Allocated->ytop(),my=ally-y-(ymin-1); y<my; ++y)
        for(unsigned x=0,mx=allx-x-(xmin-1); x<mx; ++x)
        {
        /*
            fprintf(stderr, "Testing (%u,%u) (need box of (", x,y);
            if(xsiz == nsize)fprintf(stderr, ">=");
            fprintf(stderr, "%u,", xmin);
            if(ysiz == nsize)fprintf(stderr, ">=");
            fprintf(stderr, "%u)) - (mx,my)=(%u,%u)\n", ymin, mx,my);
        */
            favors favor;
            if(xsiz==nsize || ysiz==nsize)
            {
                favor = favor_both;
                if(xsiz==nsize && ysiz!=nsize)favor = favor_x;
                if(ysiz==nsize && xsiz!=nsize)favor = favor_y;
            }
            else
                favor = favor_none;
            
            xycoord sq;            
            if(Allocated->BiggestEmptyAt(x,y, sq, xmin,ymin, favor))
            {
                if(found)
                {
                    if(sq.x == allx-x && xsiz!=nsize)continue;
                    if(sq.y == ally-y && ysiz!=nsize)continue;
                }
                
                unsigned xfree = (xsiz==nsize) ? sq.x : 1+allx-sq.x;
                unsigned yfree = (ysiz==nsize) ? sq.y : 1+ally-sq.y;
                unsigned freesize = xfree * yfree;
                    
                bool good = freesize > most;
                if(good)
                {
                /*
                    fprintf(stderr, "+++ sq = (%u,%u), free = (%u,%u), freesize %u > %u\n",
                        sq.x,sq.y,xfree,yfree,freesize,most);
                */
                    if(xsiz != nsize)sq.x = xmin;
                    if(ysiz != nsize)sq.y = ymin;
                    found = true;
                }
                /*
                else
                {
                    fprintf(stderr, "--- sq = (%u,%u), free = (%u,%u), freesize %u <= %u\n",
                        sq.x,sq.y,xfree,yfree,freesize,most);
                }
                */
                
                if(good)
                {
                    most = freesize;
                    mostpos.x = x;
                    mostpos.y = y;
                    mostsiz = sq;
                }
            }
            else
            {
                unsigned nlenx = Allocated->nanlen_x(x,y);
            /*
                fprintf(stderr, "/// %u,%u not available - skipping %u\n", xmin,ymin, nlenx);
            */
                x += nlenx-1;
            }
        }
    
    if(!most)return ScreenRegion();    
    
    return ScreenRegion(this, mostsiz, mostpos, temporary);
}

void Screen::Locate(const unsigned x, const unsigned y)
{
    if(y > ypos)
    {
        xpos = 0;
        do { Gputch('\n'); ++ypos; } while(y > ypos);
    }
    else if(y < ypos)
    {
        Gprintf("\33[%uA", ypos-y);
        ypos = y;
    }
    if(xpos != x)
    {
        if(!x) Gputch('\r');
        else if(x == xpos-1) Gputch('\b');
        else if(x == xpos-2) Gprintf("\b\b");
        else if(x < xpos)Gprintf("\33[%uD", xpos-x);
        else if(x > xpos)Gprintf("\33[%uC", x-xpos);
        xpos = x;
    }
}
void Screen::Color(BYTE attr)
{
    if(attr == textattr)return;
    bool needpp = false;
    static const char clrswap[] = "04261537";
    Gprintf("\33[");
    
#if 0&&defined(linux)
    if(!(attr&0x08) && (textattr&0x08)) { if(needpp)Gputch(';');Gprintf("21");needpp=true; }
    if(!(attr&0x80) && (textattr&0x80)) { if(needpp)Gputch(';');Gprintf("25");needpp=true; }
#else
    if((attr&8) < (textattr&8)
    || (attr&0x80) < (textattr&0x80))
    {
        if(needpp)Gputch(';');Gputch('0');needpp=true;
        textattr = (attr & 0x77) ^ 0x77;
    }
#endif    
    if((attr&0x08) && !(textattr&0x08)) { if(needpp)Gputch(';');Gputch('1');needpp=true; }
    if((attr&0x80) && !(textattr&0x80)) { if(needpp)Gputch(';');Gputch('5');needpp=true; }
    if((attr&0x07) != (textattr&0x07)) { if(needpp)Gputch(';');Gprintf("3%c",clrswap[(attr   )&7]);needpp=true; }
    if((attr&0x70) != (textattr&0x70)) { if(needpp)Gputch(';');Gprintf("4%c",clrswap[(attr>>4)&7]);needpp=true; }
    textattr = attr;
    Gputch('m');
}
void Screen::Write(unsigned x,unsigned y, BYTE attr, char c)
{
    Locate(x, y);
    Color(attr);
    Gputch(c);
    ++xpos;
}
void Screen::Write(unsigned x,unsigned y)
{
    Locate(x, y);
    Flush();
}
void Screen::Flush()
{
    fflush(stdout);
}
void Screen::UnderlineCursor()
{
#if 0&&defined(linux)
    Gprintf("\33[?2c");
    Flush();
#endif
}
void Screen::BlockCursor()
{
#if 0&&defined(linux)
    Gprintf("\33[?26;;8c");
    Flush();
#endif
}

Screen Display;
